__submodules__ = ["installer"]

# <AUTOGEN_INIT>
from kslurm.installer.installer import (
    METADATA_URL,
    install,
    run_installation,
)

__all__ = ["METADATA_URL", "install", "run_installation"]

# </AUTOGEN_INIT>



from __future__ import absolute_import, annotations

import itertools as it
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import kslurm.args.helpers as helpers
from kslurm.args.arg_sorter import ArgSorter
from kslurm.args.arg_types import Arg, FlagArg, KeywordArg, PositionalArg, TailArg
from kslurm.args.help import print_help
from kslurm.exceptions import (
    CommandLineError,
    CommandLineErrorGroup,
    MandatoryArgError,
    TailError,
)

T = TypeVar("T")
S = TypeVar("S")


def parse_args(
    args: Iterable[str],
    models: T,
    escalate_errors: bool = False,
    script_name: str = "",
    docstring: str = "",
) -> T:
    try:
        return _parse_args(args, models, script_name, docstring)
    except CommandLineError as err:
        print(err.msg)
        if isinstance(err, CommandLineErrorGroup):
            [print(e.msg) for e in err.sub_errors]
        if escalate_errors:
            raise err
        else:
            exit()


def _parse_args(args: Iterable[str], models: T, script_name: str, docstring: str) -> T:
    model_list = helpers.get_arg_list(models)
    model_list.append(FlagArg(id="help", match=["--help", "-h"]))

    sorted_models = ArgSorter(model_list)

    matchers = sorted_models.dynamic_args

    labelled = [_classify_arg(a, matchers) for a in args]
    grouped_args = list(
        _group_args(labelled, sorted_models.static_args, sorted_models.tail)
    )

    show_help = False
    for arg in grouped_args:
        if arg.id == "help":
            show_help = arg.value
            break

    updated = helpers.update_model(
        filter(lambda arg: arg.id != "help", grouped_args), models
    )

    if show_help:
        print_help(script_name, models, docstring)
        exit(0)

    sorted_args = ArgSorter(helpers.get_arg_list(updated))

    validation_errs = filter(
        None, [arg.validation_err for arg in sorted_args.validated_args]
    )

    tail_arg = sorted_args.tail
    if tail_arg and tail_arg.raise_exception:
        tail_err = [
            TailError(
                f"{tail_arg.values} does not match any Shape, Keyword, or Positional "
                "Args."
            )
        ]
    else:
        tail_err = []

    mandatory_arg_errs = cast(List[MandatoryArgError], [])
    for arg in helpers.get_arg_list(updated):
        try:
            arg.value
        except MandatoryArgError as err:
            mandatory_arg_errs.append(err)

    all_err = list(it.chain(validation_errs, tail_err, mandatory_arg_errs))

    if all_err:
        raise CommandLineErrorGroup("Problems found!", all_err)

    return updated


def _classify_arg(arg: str, arg_list: Iterable[Arg[Any]]):
    for argtype in arg_list:
        if argtype.match(arg):
            return argtype.set_value(arg)
    return cast(Arg[Any], PositionalArg().set_value(arg))


def _group_args(
    args: List[Arg[Any]],
    positional_models: Iterable[PositionalArg[Any]],
    tail_model: Optional[TailArg],
) -> Iterable[Arg[Any]]:

    group_size: int = 0
    prev_arg: Optional[Arg[Any]] = None
    positional_iter: Iterator[PositionalArg[Any]] = iter(positional_models)
    terminated = False

    updated_args: Iterable[Optional[Arg[Any]]] = []
    next_update = updated_args

    kcast: Callable[[Any], KeywordArg[Any]] = lambda x: cast(KeywordArg[Any], x)
    for arg in args:
        # Pyright is stupid and adds Keyword[Unknown] as a type annotation following
        # isinstance(_, KeywordArg). This messes up all the following code so we make a
        # a new variable to bear the burden of the unnecessary annotation and avoid
        # lots of casting
        prev_kw = prev_arg
        if (
            group_size
            and isinstance(prev_kw, KeywordArg)
            and (isinstance(arg, PositionalArg) or not prev_kw.lazy)
        ):
            prev_arg = kcast(
                prev_kw.add_values(it.chain(prev_kw.values, [arg.raw_value]))
            )
            group_size -= 1
            next_update = it.chain(updated_args, [prev_arg])
            continue

        updated_args = next_update

        if terminated:
            prev_arg, tail_model = _process_tail(arg, tail_model, True)
            group_size = _get_keyword_group_size(prev_arg)

        elif isinstance(arg, PositionalArg):
            new_arg, tail_model = _process_positional(arg, positional_iter, tail_model)
            if new_arg is None:
                new_arg = prev_arg
            group_size = _get_keyword_group_size(new_arg)
            prev_arg = new_arg

        else:
            group_size = _get_keyword_group_size(arg)
            prev_arg = arg

        if prev_arg and prev_arg.terminal:
            terminated = True

        next_update = it.chain(updated_args, [prev_arg])

    updated_args = next_update

    return list(filter(None, updated_args))


def _get_keyword_group_size(arg: Optional[Arg[Any]]):
    if isinstance(arg, KeywordArg):
        return arg.num
    else:
        return 0


def _process_positional(
    arg: PositionalArg[Any],
    positional_models: Iterator[PositionalArg[Any]],
    tail_model: Optional[TailArg],
    terminate: bool = True,
) -> Tuple[Union[PositionalArg[Any], TailArg, None], Optional[TailArg]]:
    try:
        model = next(positional_models)
    except StopIteration:
        return _process_tail(arg, tail_model, terminate)
    return model.set_value(arg.raw_value), tail_model


def _process_tail(
    arg: Arg[Any], tail_model: Optional[TailArg], terminate: bool = False
):
    if tail_model:
        new_tail = tail_model.add_values([arg.raw_value])
    else:
        ret = TailArg().add_values([arg.raw_value])
        ret.raise_exception = True
        new_tail = ret

    if terminate:
        return new_tail, None
    return None, new_tail



from __future__ import absolute_import

import json
from collections import UserDict
from pathlib import Path


class KpyIndex(UserDict[str, str]):
    def __init__(self, slurm_tmpdir: Path):
        self._path = slurm_tmpdir / "tmp" / "kpy-index.json"
        if not self._path.exists():
            self._path.parent.mkdir(parents=True, exist_ok=True)
            with self._path.open("w") as f:
                json.dump({}, f)
                self.data = {}
        else:
            with self._path.open("r") as f:
                self.data = json.load(f)

    def write(self):
        with self._path.open("w") as f:
            json.dump(self.data, f)



from __future__ import absolute_import

import attr

from kslurm.args import TailArg
from kslurm.args.arg_types import SubCommand
from kslurm.args.command import command
from kslurm.cli import kbatch, kjupyter, kpy, krun
from kslurm.cli.config import config
from kslurm.installer import install

NAME = "kslurm"
HOME_DIR = "KSLURM_HOME"
ENTRYPOINTS = ["kbatch", "krun", "kjupyter", "kslurm", "kpy"]


@attr.s(auto_attribs=True)
class KslurmModel:
    command: SubCommand = SubCommand(
        commands={
            "kbatch": kbatch,
            "krun": krun,
            "kjupyter": kjupyter,
            "kpy": kpy,
            "config": config,
            "update": lambda x: None,  # type: ignore
        },
    )

    tail: TailArg = TailArg("Args")


@command
def main(args: KslurmModel) -> None:
    command = args.command.value
    tail = args.tail
    name = f"kslurm {args.command.raw_value}"
    if args.command.raw_value == "update":
        install(tail.values, NAME, HOME_DIR, ENTRYPOINTS)
    command([name, *tail.values])


if __name__ == "__main__":
    main(["kslurm", "kbatch", "-h"])



from __future__ import absolute_import

import hashlib
import importlib.resources as impr
import os
import re
import shutil
import subprocess as sp
import tarfile
import tempfile
from collections import UserDict
from pathlib import Path
from typing import Any, Literal, Optional, Union, overload

import attr
import virtualenv  # type: ignore
from virtualenv.create import pyenv_cfg  # type: ignore

from kslurm.appconfig import get_config
from kslurm.args.arg_types import (
    FlagArg,
    KeywordArg,
    PositionalArg,
    ShapeArg,
    SubCommand,
    TailArg,
)
from kslurm.args.command import command
from kslurm.kpyindex import KpyIndex
from kslurm.shell import Shell


def get_hash(item: Union[str, bytes]):
    if isinstance(item, str):
        item = item.encode()
    return hashlib.md5(item).hexdigest()


def pip_freeze(venv_dir: Path):
    return sp.run(
        [venv_dir / "bin" / "python", "-m", "pip", "freeze"], capture_output=True
    ).stdout


def _get_unique_name(index: KpyIndex, stem: str = "venv", i: int = 0) -> str:
    if i == 0:
        candidate = stem
    else:
        candidate = f"{stem}{i}"
    if candidate in index:
        return _get_unique_name(index, stem, i + 1)
    return candidate


class MissingPipdirError(Exception):
    def __init__(self, msg: str, *args: Any):
        super().__init__(msg, *args)
        self.msg = msg


class MissingSlurmTmpdirError(Exception):
    def __init__(self, msg: str, *args: Any):
        super().__init__(msg, *args)
        self.msg = msg


@overload
def _get_slurm_tmpdir(allow_missing: Literal[True] = ...) -> Optional[Path]:
    ...


@overload
def _get_slurm_tmpdir(allow_missing: Literal[False] = ...) -> Path:
    ...


def _get_slurm_tmpdir(allow_missing: bool = True):
    if not os.environ.get("SLURM_TMPDIR"):
        if not allow_missing:
            raise MissingSlurmTmpdirError(
                "This command can only be used in a compute node. Use `krun` to start "
                "an interactive session"
            )
        return
    return Path(os.environ["SLURM_TMPDIR"])


class VenvCache(UserDict[str, Path]):
    def __init__(self):
        pipdir = get_config("pipdir")
        if pipdir is None:
            raise MissingPipdirError(
                "pipdir not set. Please set pipdir using `kslurm config pipdir "
                "<directory>`, typically to a project-space or permanent storage "
                "directory"
            )
        self.venv_cache = Path(pipdir, "venv_archives")
        self.venv_cache.mkdir(exist_ok=True)
        venvs_re = [
            re.search(r"(.+)\.tar\.gz$", str(f.name)) for f in self.venv_cache.iterdir()
        ]
        venvs = [r.group(1) for r in venvs_re if r]
        self.data = {v: self._construct_path(v) for v in venvs}

    def get_path(self, name: str):
        try:
            return self.data[name]
        except KeyError:
            return self._construct_path(name)

    def _construct_path(self, name: str):
        return (self.venv_cache / name).with_suffix(".tar.gz")


@command
def _bash():
    """Echo script for inclusion in .bashrc

    e.g.
        kpy bash >> $HOME/.bashrc
    """
    with impr.path("kslurm.bin", "bash.sh") as path:
        print(f"\nsource {path.resolve()}")


@attr.frozen
class _LoadModel:
    name: PositionalArg[str] = PositionalArg(value="")
    new_name: KeywordArg[str] = KeywordArg(match=["--as"])


@command
def _load(args: _LoadModel):
    slurm_tmp = _get_slurm_tmpdir()
    if slurm_tmp:
        index = KpyIndex(slurm_tmp)
        name = args.name.value

        label = args.new_name.values[0] if args.new_name.values else name
        if label in index:
            print(
                f"An environment called '{label}' already exists. You can load "
                f"'{name}' under a different name using --as:\n"
                f"\tkpy load {label} --as <name>\n"
                f"You can also activate the existing '{label}' using\n"
                f"\tkpy activate {label}"
            )
            return
        venv_dir = Path(tempfile.mkdtemp(prefix="kslurm-venv-", dir=slurm_tmp / "tmp"))
    else:
        index = None
        name = args.name.value
        label = name
        venv_dir = Path(tempfile.mkdtemp(prefix="kslurm-"))

    pipdir = get_config("pipdir")
    if not pipdir:
        print(
            "pipdir not set. Please set pipdir using `kslurm config pipdir "
            "<directory>`, typically to a project-space or permanent storage directory"
        )
        return

    name = args.name.value

    try:
        venv_cache = VenvCache()
    except MissingPipdirError as err:
        print(err.msg)
        return

    if not name or name not in venv_cache:
        print("Valid venvs:\n\t" + "\n\t".join(venv_cache))
        return

    # pyload_venv_dir = slurm_tmpdir / "__virtual_environments__"
    # if (pyload_venv_dir / name).exists():
    #     shell = Shell.get()
    #     shell.activate(pyload_venv_dir / name)

    print(f"Unpacking venv '{name}'", end="")
    if label != name:
        print(f" as '{label}'")
    else:
        print()
    with tarfile.open(venv_cache[name], "r") as tar:
        tar.extractall(venv_dir)

    print("Updating paths")
    with (venv_dir / "bin" / "activate").open("r") as f:
        lines = f.read().splitlines()
        # \x22 and \x27 are " and ' char
        subbed = [
            re.sub(
                r"(?<=^VIRTUAL_ENV=([\x22\x27])).*(?=\1$)",
                str(venv_dir),
                line,
            )
            for line in lines
        ]
    with (venv_dir / "bin" / "activate").open("w") as f:
        f.write("\n".join(subbed))

    for root, _, files in os.walk(venv_dir / "bin"):
        for file in files:
            path = Path(root, file)
            if path.is_symlink() or not os.access(path, os.X_OK):
                continue
            with path.open("r") as f:
                lines = f.read().splitlines()
                subbed = [
                    re.sub(
                        r"(?<=^#!)/.*$",
                        str(venv_dir / "bin" / "python"),
                        line,
                    )
                    for line in lines
                ]
            with path.open("w") as f:
                f.write("\n".join(subbed))

    cfg: Any = pyenv_cfg.PyEnvCfg.from_folder(venv_dir)  # type: ignore
    cfg.update({"prompt": label, "state_hash": get_hash(pip_freeze(venv_dir))})
    cfg.write()

    if index is not None:
        index[label] = str(venv_dir)
        index.write()

    shell = Shell.get()
    if n := shell.activate(venv_dir):
        kpy(n)


@attr.frozen
class _SaveModel:
    name: PositionalArg[str] = PositionalArg()
    force: FlagArg = FlagArg(match=["--force", "-f"])


@command
def _save(args: _SaveModel):
    if not os.environ.get("VIRTUAL_ENV"):
        print(
            "No active virtual env detected. Please activate one, or ensure "
            "$VIRTUAL_ENV is being set correctly"
        )
    try:
        venv_cache = VenvCache()
    except MissingPipdirError as err:
        print(err.msg)
        return

    name = args.name.value
    delete = False
    if name in venv_cache:
        if args.force.value:
            delete = True
        else:
            print(f"{name} already exists. Run with -f to force overwrite")
            return

    dest = venv_cache.get_path(name)

    _, tmp = tempfile.mkstemp(prefix="kslurm-", suffix="tar.gz")

    venv_dir = Path(os.environ["VIRTUAL_ENV"])
    cfg: Any = pyenv_cfg.PyEnvCfg.from_folder(venv_dir)  # type: ignore
    cfg.update(
        {
            "prompt": args.name.value,
            "state_hash": get_hash(pip_freeze(venv_dir)),
        }
    )
    cfg.write()
    with tarfile.open(tmp, mode="w:gz") as tar:
        tar.add(venv_dir, arcname="")

    if delete:
        os.remove(dest)
    slurm_tmp = _get_slurm_tmpdir()
    if slurm_tmp:
        index = KpyIndex(slurm_tmp)
        index[args.name.value] = str(venv_dir)
        index.write()
    shutil.move(tmp, dest)


@attr.frozen
class _CreateModel:
    name: PositionalArg[str] = PositionalArg("")
    version: ShapeArg[str] = ShapeArg(
        value="", match=lambda s: bool(re.match(r"^[23]\.\d{1,2}$", s))
    )


@command
def _create(args: _CreateModel):
    if args.version.value:
        ver = ["-p", args.version.value]
    else:
        ver = []

    slurm_tmp = _get_slurm_tmpdir()
    if slurm_tmp:
        index = KpyIndex(slurm_tmp)
        name = args.name.value if args.name.value else _get_unique_name(index, "venv")
        if args.name.value in index:
            print(
                f"An environment called '{name}' already exists. You can activate "
                f"the existing '{name}' using\n"
                "\tkpy activate {name}"
            )
            return

        venv_dir = tempfile.mkdtemp(prefix="kslurm-venv-", dir=slurm_tmp / "tmp")
        no_download = ["--no-download"]
        no_index = ["--no-index"]
    else:
        index = None
        name = args.name.value if args.name.value else "venv"
        venv_dir = tempfile.mkdtemp(prefix="kslurm-")
        no_download = []
        no_index = []

    try:
        virtualenv.cli_run(  # type: ignore
            [venv_dir, "--symlinks", "--prompt", name, *ver, *no_download]
        )
        sp.run(
            [
                os.path.join(venv_dir, "bin", "python"),
                "-m",
                "pip",
                "install",
                "--upgrade",
                "pip",
                *no_index,
            ]
        )
    except RuntimeError as err:
        print(err.args[0])
        return
    if index is not None:
        index[name] = str(venv_dir)
        index.write()

    shell = Shell.get()
    n = shell.activate(Path(venv_dir))
    if n:
        kpy(n)


@attr.frozen
class ActivateModel:
    name: PositionalArg[str] = PositionalArg()


@command
def _activate(args: ActivateModel):
    try:
        slurm_tmp = _get_slurm_tmpdir(False)
    except MissingSlurmTmpdirError as err:
        print(err.msg)
        return

    index = KpyIndex(slurm_tmp)
    name = args.name.value
    if name not in index:
        print(
            f"An environment with the name '{name}' has not yet been initialized. ",
            end="",
        )
        try:
            venv_cache = VenvCache()
            if name in venv_cache:
                print(
                    f"The saved environment called '{name}' can be loaded using\n"
                    f"\tkpy load {name}\n"
                )
        except MissingPipdirError:
            pass
        print(f"A new environment can be created using\n\tkpy create {name}")
        return

    shell = Shell.get()
    n = shell.activate(Path(index[name]))
    if n:
        kpy(n)


@command
def _list():
    try:
        slurm_tmp = _get_slurm_tmpdir(False)
    except MissingSlurmTmpdirError as err:
        print(err.msg)
        return
    index = KpyIndex(slurm_tmp)
    print("\n".join(index))


@command
def _refresh():
    try:
        dir = Path(os.environ["VIRTUAL_ENV"])
        cfg: Any = pyenv_cfg.PyEnvCfg.from_folder(dir)  # type: ignore
        state_hash = cfg["state_hash"]
    except KeyError:
        return

    try:
        hsh = get_hash(pip_freeze(dir))
        if hsh != state_hash and cfg["prompt"][0] != "*":
            cfg["prompt"] = "*" + cfg["prompt"]
        elif hsh == state_hash and cfg["prompt"][0] == "*":
            cfg["prompt"] = cfg["prompt"][1:]
        cfg.write()
    except KeyError:
        return


@attr.frozen
class KpyModel:
    command: SubCommand = SubCommand(
        commands={
            "load": _load,
            "save": _save,
            "bash": _bash,
            "create": _create,
            "activate": _activate,
            "list": _list,
            "_refresh": _refresh,
        },
    )

    tail: TailArg = TailArg("Args")


@command
def kpy(args: KpyModel) -> None:
    command = args.command.value
    tail = args.tail
    name = f"kslurm {args.command.raw_value}"
    command([name, *tail.values])


if __name__ == "__main__":
    kpy(["kpy", "bash", "--help"])



from __future__ import absolute_import

import attr

import kslurm.appconfig as appconfig
from kslurm.args.arg_types import FlagArg, PositionalArg
from kslurm.args.command import command


@attr.frozen
class ConfigModel:
    entry: PositionalArg[str] = PositionalArg(help="Configuration value to update")
    value: PositionalArg[str] = PositionalArg(
        value="",
        help="Updated value for config entry. If not provided, the current config "
        "value is printed",
    )
    help: FlagArg = FlagArg(match=["-h", "--help"], help="Show this help message")


@command
def config(args: ConfigModel):
    if not args.value.value:
        value = appconfig.get_config(args.entry.value)
        if value is None:
            print("")
        else:
            print(value)

    else:
        appconfig.set_config(args.entry.value, args.value.value)



from __future__ import absolute_import, annotations

import abc
import copy
from typing import Callable, Dict, Generic, Iterable, List, Optional, TypeVar

from colorama import Fore, Style

from kslurm.args.types import WrappedCommand
from kslurm.exceptions import MandatoryArgError, ValidationError

T = TypeVar("T")
S = TypeVar("S")


class Arg(abc.ABC, Generic[T]):
    raw_value: str

    def __init__(
        self,
        *,
        id: str,
        match: Callable[[str], bool],
        value: Optional[str],
        format: Callable[[str], T],
        help: str,
    ):
        self.id = id
        self.match = match
        self._format = format
        self.value = value
        self.help = help
        self.terminal: bool = False

    @property
    def value(self):
        if self._value is None:
            raise MandatoryArgError(f"{self.name} has not been provided a value.")
        return self._value

    @value.setter
    def value(self, value: Optional[str]):
        if value is None:
            self._value = value
            self.raw_value = ""
        else:
            self._value = self._format(value)
            self.raw_value = value

    def __eq__(self, o: object) -> bool:
        if isinstance(o, self.__class__):
            return o._value == self._value
        else:
            return False

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.id} = {self._value}>"

    def __str__(self) -> str:
        return str(self.value)

    def __hash__(self) -> int:
        return hash(hash(self.id) + hash(self.value))

    def setid(self, id: str):
        c = copy.copy(self)
        c.id = id
        return c

    @property
    def name(self) -> str:
        return self.id

    @abc.abstractmethod
    def set_value(self, value: str) -> Arg[T]:
        pass


class PositionalArg(Arg[T]):
    def __init__(
        self,
        value: Optional[str] = None,
        id: str = "positional",
        format: Callable[[str], T] = str,
        validator: Callable[[str], str] = lambda x: x,
        help: str = "",
        name: str = "",
    ):
        super().__init__(
            id=id, match=lambda x: True, value=value, format=format, help=help
        )
        self.validator = validator
        self._name = name
        self.updated = False
        self.validation_err: Optional[ValidationError] = None

    def set_value(self, value: str):
        c = copy.copy(self)
        try:
            c.value = self.validator(value)
            c.updated = True
            return c
        except ValidationError as err:
            c.validation_err = ValidationError(
                f"""
    {Fore.RED + Style.BRIGHT}ERROR:{Style.RESET_ALL}
        Invalid value for "{Style.BRIGHT + self.name + Style.RESET_ALL}":
            {err.msg}
            """
            )
            return c

    @property
    def name(self):
        if self._name:
            return self._name
        return self.id


class ChoiceArg(PositionalArg[T]):
    def __init__(
        self,
        *,
        value: Optional[str] = None,
        match: List[str],
        id: str = "",
        format: Callable[[str], T] = str,
        help: str = "",
        name: str = "",
    ):
        def check_match(val: str):
            if val in match:
                return val
            choices = "\n".join([f"\t\t• {m}" for m in match])
            raise ValidationError(f"Please select between:\n" f"{choices}")

        super().__init__(
            value=value,
            id=id,
            format=format,
            validator=check_match,
            help=help,
            name=name,
        )

        self.match_list = match


class SubCommand(ChoiceArg[WrappedCommand]):
    def __init__(
        self,
        *,
        commands: Dict[str, WrappedCommand],
        value: Optional[str] = None,
        id: str = "",
    ):
        self.commands = commands
        super().__init__(
            value=value,
            id=id,
            format=lambda s: self.commands[s],
            match=list(commands.keys()),
            help="Run any command followed by -h for more inforamtion",
            name="Command",
        )
        self.terminal = True


class ShapeArg(Arg[T]):
    def __init__(
        self,
        *,
        id: str = "",
        match: Callable[[str], bool],
        value: Optional[str] = None,
        format: Callable[[str], T] = str,
        help: str = "",
        name: str = "",
        syntax: str = "",
        examples: List[str] = [],
    ):
        super().__init__(id=id, match=match, value=value, format=format, help=help)
        self.updated = False
        self._name = name
        self.syntax = syntax
        self.examples = examples

    def set_value(self, value: str):
        c = copy.copy(self)
        c.value = value
        c.raw_value = value
        c.updated = True
        return c

    @property
    def name(self):
        if self._name:
            return self._name
        return self.id


class FlagArg(Arg[bool]):
    def __init__(
        self,
        *,
        id: str = "",
        match: List[str],
        value: Optional[bool] = False,
        help: str = "",
    ):
        def check_match(val: str):
            if val in match:
                return True
            return False

        if value:
            raw_value = match[0]
        elif value is None:
            raw_value = None
        else:
            raw_value = ""

        super().__init__(
            id=id, match=check_match, format=bool, value=raw_value, help=help
        )
        self.match_list = match

    def set_value(self, value: str):
        c = copy.copy(self)
        c.value = value
        return c

    @property
    def name(self):
        return ", ".join(self.match_list)


class KeywordArg(FlagArg, Generic[S]):
    def __init__(
        self,
        *,
        id: str = "",
        match: List[str],
        value: Optional[bool] = False,
        num: int = 1,
        validate: Callable[[str], str] = lambda x: x,
        values: List[str] = [],
        help: str = "",
        values_name: str = "",
        lazy: bool = False,
    ):
        super().__init__(id=id, match=match, help=help, value=value)
        self.num = num
        self.validate = validate
        self.values = values
        self.values_name = values_name
        self.validation_err: Optional[ValidationError] = None
        self.lazy = lazy

    @property
    def values(self) -> List[str]:
        return self._values

    @values.setter
    def values(self, values: List[str]):
        self._values = values

    def set_value(self, value: str):
        c = copy.copy(self)
        c.value = value
        # We need to obliterate any values in order for argparse to work
        # Otherwise any default args will be included in the final list.
        # A more elegant solution would delete default args organically
        # if user-supplied args are added.
        c.values = []
        return c

    def add_values(self, values: Iterable[str]):
        c = copy.copy(self)
        try:
            c.values = list(map(self.validate, values))
            return c
        except ValidationError as err:
            c.validation_err = ValidationError(
                f"""
    {Fore.RED + Style.BRIGHT}ERROR:{Style.RESET_ALL}
        Invalid value for "{Style.BRIGHT + self.id + Style.RESET_ALL}":
            {err.msg}
            """,
            )
            return c

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.id} = {self.values}>"

    def __str__(self) -> str:
        return " ".join(self.values)

    def __eq__(self, o: object) -> bool:
        if isinstance(o, self.__class__):
            return o._value == self._value and o.values == self.values
        else:
            return False

    def __hash__(self) -> int:
        value_hash = hash(sum([hash(val) for val in self.values]))
        return hash(super().__hash__() + value_hash)


class TailArg(KeywordArg[str]):
    def __init__(self, name: str = "Tail"):
        super().__init__(id="tail", num=-1, match=[])
        self._name = name
        self.raise_exception = False

    @property
    def name(self):
        return self._name



# pyright: reportUnknownMemberType=false, reportUnknownArgumentType=false
from __future__ import absolute_import

from typing import Any, List, cast

from kslurm.args import Arg
from kslurm.args.arg_types import KeywordArg
from kslurm.args.helpers import get_arg_list


class ModelTest:
    def __init__(
        self,
        model: object,
        tests: List[List[str]],
        expected_outcomes: List[List[Arg[Any]]],
    ):
        self.model = model
        self.tests = tests
        self.expected_outcomes = expected_outcomes


def update_model(models: object, updates: List[Any]):
    arg_list = get_arg_list(models)
    ret = []
    for model, update in zip(arg_list, updates):
        if isinstance(update, list):
            val, values = update  # type: ignore
            assert isinstance(model, KeywordArg)
            if val is None:
                ret.append(model.add_values(values))
            else:
                ret.append(model.set_value(str(val)).add_values(values))
        elif update is None:
            ret.append(model)
        else:
            ret.append(model.set_value(str(update)))
    return cast(List[Arg[Any]], ret)



from __future__ import absolute_import


def time(time: str):
    if ":" in time:
        if "-" in time:
            days, hhmm = time.split("-")
            hours, min = hhmm.split(":")
        else:
            days = 0
            hours, min = time.split(":")
    else:
        try:
            min = int(time)
            hours, days = 0, 0
        except Exception:
            raise TypeError(
                f'Invalid format for time: "{time}"\n'
                f"Must be as [xx-]xx:xx or x where x is a number"
            )
    return str(int(min) + int(hours) * 60 + int(days) * 60 * 24)



from __future__ import absolute_import

import subprocess
from typing import Union

import kslurm.text as txt
from kslurm.args.command import command
from kslurm.exceptions import TemplateError
from kslurm.models import SlurmModel
from kslurm.slurm import SlurmCommand


@command
def krun(args: Union[SlurmModel, TemplateError]):
    """Start an interactive session

    If no command is provided, an interactive session will begin. The server name
    on your shell prompt will change to the name of your cluster, and you can
    run any job you wish from there. To exit the session, enter "exit".

    If a command is provided, it will be immediately submitted using srun. The output
    of the command will be displayed on the console.

    Using this command, jobs will begin much quicker than sbatch, however, it should
    only be used for interactive sessions. Note that you should request jobs no longer
    than 3hr. Anything longer, and it could take a very long time for the job to begin.

    When your command contains bash interpreted elements such as $VARIABLES and
    $(subshells), these will be immediately expanded. Normally, this behaviour
    is fine, but sometimes they should only be interpretted on the allocated cluster.
    For instance, $SLURM_TMPDIR will evaluate to "" unless it is interpretted later.
    To force this behaviour, wrap the $VARIABLE or $(subshell) in quotes:
        '$SLURM_TMPDIR'
        '$(hostname)'
    """
    slurm = SlurmCommand(args)
    if slurm.command:
        print(txt.KRUN_CMD_MESSAGE.format(args=slurm.slurm_args, command=slurm.command))
    else:
        print(txt.INTERACTIVE_MSG.format(args=slurm.slurm_args))

    if not slurm.test:
        subprocess.run(slurm.run, shell=True)



__submodules__ = ["args", "installer", "models", "slurm"]

__ignore__ = ["main"]

# <AUTOGEN_INIT>
from kslurm.args import (
    Arg,
    ArgSorter,
    ChoiceArg,
    FlagArg,
    KeywordArg,
    PositionalArg,
    ShapeArg,
    SubCommand,
    TailArg,
    group_by_type,
    parse_args,
    print_help,
)
from kslurm.installer import (
    METADATA_URL,
    install,
    run_installation,
)
from kslurm.models import (
    SlurmModel,
    TemplateArgs,
    Templates,
    UpdateModel,
    VERSION_REGEX,
    job_template,
    list_templates,
    mem,
    set_template,
    templates,
    time,
)
from kslurm.slurm import (
    SlurmCommand,
    div_remainder,
    slurm_time_format,
)

__all__ = [
    "Arg",
    "ArgSorter",
    "ChoiceArg",
    "FlagArg",
    "KeywordArg",
    "METADATA_URL",
    "PositionalArg",
    "ShapeArg",
    "SlurmCommand",
    "SlurmModel",
    "SubCommand",
    "TailArg",
    "TemplateArgs",
    "Templates",
    "UpdateModel",
    "VERSION_REGEX",
    "div_remainder",
    "group_by_type",
    "install",
    "job_template",
    "list_templates",
    "mem",
    "parse_args",
    "print_help",
    "run_installation",
    "set_template",
    "slurm_time_format",
    "templates",
    "time",
]

# </AUTOGEN_INIT>



from __future__ import absolute_import

import itertools as it
from collections import defaultdict
from typing import Any, DefaultDict, Dict, ItemsView, Iterable, List, TypeVar, cast

import attr

from .arg_types import Arg

T = TypeVar("T")
C = TypeVar("C", bound=Arg[Any])


def group_by_type(items: Iterable[T]) -> Dict[type, List[T]]:
    groupedDict = cast(DefaultDict[type, List[T]], defaultdict(list))
    for key, value in it.groupby(items, type):
        groupedDict[key] += list(value)
    return groupedDict


def get_arg_list(models: object):
    if hasattr(models, "__attrs_attrs__"):
        return [
            cast(Arg[Any], arg.setid(name))
            for name, arg in attr.asdict(models, recurse=False).items()
        ]
    elif isinstance(models, dict):
        return [
            arg.setid(name)
            for name, arg in cast(ItemsView[str, Arg[Any]], models.items())
        ]
    else:
        raise Exception(f"{type(models)} is not a supported object for arg models")


def update_model(arg_updates: Iterable[Arg[Any]], model: T) -> T:
    update = {arg.id: arg for arg in arg_updates}
    if hasattr(model, "__attrs_attrs__"):
        return model.__class__(**update)
    elif isinstance(model, dict):
        return cast(T, {**model, **update})
    else:
        raise Exception(f"{type(model)} is not a supported object for arg models")


def get_help(model: Any) -> bool:
    if attr.has(model):
        help = getattr(model, "help", None)
        return getattr(help, "value", False)
    if isinstance(model, dict):
        help: Any = model.get("help", None)  # type: ignore
        return getattr(help, "value", False)
    raise Exception(f"{type(model)} is not a supported type for arg models")



from __future__ import absolute_import

import re
from pathlib import Path

import attr

import kslurm.models.formatters as formatters
import kslurm.models.validators as validators
from kslurm.args import FlagArg, KeywordArg, ShapeArg, TailArg


@attr.s(auto_attribs=True)
class SlurmModel:
    time: ShapeArg[int] = ShapeArg[int](
        match=lambda arg: bool(re.match(r"^([0-9]{1,2}-)?[0-9]{1,2}:[0-9]{2}$", arg)),
        format=formatters.time,
        value="03:00",
        name="Time",
        syntax="[d-]dd:dd",
        help="Amount of time requested. Written as [days-]hr:min.",
        examples=["3:00 (3hr)", "1-00:00 (1 day)"],
    )

    gpu: FlagArg = FlagArg(match=["gpu"], help="Request a gpu instance")

    cpu: ShapeArg[int] = ShapeArg[int](
        match=lambda arg: bool(re.match(r"^[0-9]+$", arg)),
        value="1",
        format=int,
        name="Number of CPUs",
        syntax="d",
        help="Number of CPUs to request. Maximum of 32 on Graham Nodes.",
        examples=["1 (1 cpu)", "16 (16 cpus)"],
    )

    mem: ShapeArg[int] = ShapeArg[int](
        match=lambda arg: bool(re.match(r"^[0-9]+[MG]B?$", arg)),
        format=formatters.mem,
        value="4G",
        name="Memory",
        syntax="d(M|G)[B]",
        examples=["3000MB (3 GB)", "16G (16GB)"],
        help="Amount of memory to request",
    )

    x11: FlagArg = FlagArg(
        match=["x11", "--x11"],
        help="Request X-forwarding for GUI applications. Must also request "
        "X-forwarding on your SSH connection.",
    )

    directory: ShapeArg[Path] = ShapeArg[Path](
        match=lambda arg: Path(arg).exists() and Path(arg).is_dir(),
        format=Path,
        value=".",
        name="Directory",
        syntax="/absolute/path | ./relative/path",
        help="Change working directory before submitting the command. All other "
        "relative paths will be evaluated relative to the new directory.",
    )

    test: FlagArg = FlagArg(
        match=["-t", "--test"],
        help="Test mode. Will echo SLURM settings and the command without submitting "
        "anything.",
    )

    job_template: KeywordArg[str] = KeywordArg[str](
        match=["-j", "--job-template"],
        validate=validators.job_template,
        num=1,
        values_name="template",
        help="Set CPU, memory, and time from a template. "
        "Run with -J to see list of templates. Any "
        "template item can be overridden by supplying "
        "the appropriate arg.",
    )

    list_job_templates: FlagArg = FlagArg(
        match=["-J", "--list", "-l"], help="List all available templates."
    )

    account: KeywordArg[str] = KeywordArg[str](
        match=["-a", "--account"],
        num=1,
        help="Compute account to submit the job under.",
        values_name="name",
    )

    help: FlagArg = FlagArg(match=["-h", "--help"], help="Print this help message")

    tail: TailArg = TailArg("Command")



from __future__ import absolute_import

from typing import Any, List

from kslurm.args.arg_types import Arg, PositionalArg

from .arg_templates import Templates


class DummyArgs:
    positionals: List[PositionalArg[Any]] = [
        Templates.positional.set_value("Hello"),
        Templates.positional2.set_value("beautiful"),
        Templates.positional3.set_value("world"),
    ]

    extra_positionals: List[PositionalArg[Any]] = [
        Templates.positional.set_value("Hello"),
        Templates.positional2.set_value("beautiful"),
        Templates.positional3.set_value("world"),
        Templates.positional4.set_value("Here"),
        Templates.positional5.set_value("Be"),
        Templates.positional3.set_value("Extra!"),
    ]

    ###
    # Args used to test keyword groupping
    two_keyword: List[Arg[Any]] = [
        Templates.time,
        Templates.gpu,
        Templates.job_template,
        Templates.positional,
        Templates.time,
        Templates.length_5_keyword,
        Templates.positional,
        Templates.time,
        Templates.gpu,
        Templates.positional,
        Templates.positional,
        Templates.gpu,
        Templates.number,
        Templates.time,
    ]

    positional_patchwork: List[Arg[Any]] = [
        Templates.time,
        Templates.positional,
        Templates.number,
        Templates.positional2,
        Templates.number,
        Templates.positional3,
        Templates.gpu,
    ]

    positional_wrapped_keyword: List[Arg[Any]] = [
        Templates.time,
        Templates.positional,
        Templates.job_template,
        Templates.positional,
        Templates.positional,
    ]

    lazy_keywords: List[Arg[Any]] = [
        Templates.time,
        Templates.positional,
        Templates.lazy_inf_keyword,
        Templates.positional,
        Templates.positional,
        Templates.lazy_inf_keyword,
        Templates.positional,
        Templates.positional,
        Templates.positional,
        Templates.positional,
        Templates.time,
    ]

    greedy_keywords: List[Arg[Any]] = [
        Templates.time,
        Templates.gpu,
        Templates.greedy_inf_keyword,
        Templates.positional,
        Templates.time,
        Templates.lazy_inf_keyword,
        Templates.time,
        Templates.positional,
        Templates.time,
    ]


class ArgStr:
    positionals: List[str] = ["Hello", "beautiful", "world"]

    extra_positionals: List[str] = [
        "Hello",
        "beautiful",
        "world",
        "Here",
        "be",
        "extra",
    ]

    two_keyword: List[str] = [
        "07:23",
        "gpu",
        "--job-template",
        "parcour",
        "3",
        "--length_5_keyword",
        "one",
        "of",
        "the",
        "five",
        "values",
        "command",
    ]

    shapes_and_positionals: List[str] = [
        "07:23",
        "gpu",
        "--job-template",
        "parcour",
        "one",
        "of" "3",
        "the",
        "five",
        "values",
        "command",
    ]



from __future__ import absolute_import

import re

import attr

from kslurm.args import ShapeArg, TailArg

# v(major_version).(minor_version).(patch_version).(sub_patch_version)
VERSION_REGEX = (
    r"v?(\d+)(?:\.(\d+))?(?:\.(\d+))?(?:\.(\d+))?"
    "("
    "[._-]?"
    r"(?:(stable|beta|b|rc|RC|alpha|a|patch|pl|p)((?:[.-]?\d+)*)?)?"
    "([.-]?dev)?"
    ")?"
    r"(?:\+[^\s]+)?"
)


@attr.s(auto_attribs=True)
class UpdateModel:
    version: ShapeArg[str] = ShapeArg[str](
        match=lambda arg: bool(re.match(VERSION_REGEX, arg)),
        value="",
        name="Version",
        syntax="d[.d][.d][.d][_tag]",
        examples=["0.1.0"],
        help="Optional: version to install",
    )

    tail: TailArg = TailArg()



from __future__ import absolute_import

import re
import signal
import subprocess
import sys
from typing import Union

import kslurm.text as txt
from kslurm.args.command import command
from kslurm.exceptions import TemplateError
from kslurm.models import SlurmModel
from kslurm.slurm import SlurmCommand
from kslurm.style.console import console


@command
def kjupyter(args: Union[SlurmModel, TemplateError]):
    """Start a Jupyter session

    Begins a Jupyter session on an interactive node. By default, it requests 3hr. No
    more time than this should be requested, as this could significantly delay the
    start of the server.

    For the command to work, a virtualenv containing jupyter-lab should already be
    activated. Use `pip install jupyter-lab`
    """
    slurm = SlurmCommand(args)

    slurm.command = ["jupyter-lab", "--ip", "$(hostname -f)", "--no-browser"]
    print(txt.KRUN_CMD_MESSAGE.format(args=slurm.slurm_args, command=slurm.command))

    if not slurm.test:

        proc = subprocess.Popen(
            slurm.run,
            shell=True,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        jobid = "0"

        def signal_handler(*args):  # type: ignore
            subprocess.run(f"scancel {jobid}", shell=True)
            sys.exit(0)

        signal.signal(signal.SIGINT, signal_handler)  # type: ignore

        while proc.poll() is None:
            if proc.stdout:
                line = proc.stdout.readline().decode().strip()
                queued = re.match(
                    r"^srun: job (\d+) queued and waiting for resources", line
                )
                url = re.match(
                    r"^https?:\/\/((?:[\w-]+\.)+(?:[\w]+)):(\d+)(\/lab\?token=\w+)$",
                    line,
                )
                if queued:
                    jobid = queued.group(1)
                elif any(
                    [
                        re.match(r"^srun: job \d+ has been allocated resources", line),
                        re.match(
                            r"^To access the server, open this file in a browser:$",
                            line,
                        ),
                        re.match(r"^file:\/\/\/(?:[\w.\-_]+\/)+[\w.\-_]+\.html$", line),
                        re.match(r"^Or copy and paste one of these URLs:$", line),
                        re.match(
                            r"^or http:\/\/(?:[\d]+\.)+\d+:\d+\/lab\?token=\w+$", line
                        ),
                    ]
                ):
                    pass
                elif url:
                    domain = url.group(1)
                    port = url.group(2)
                    path = url.group(3)
                    console.print(
                        txt.JUPYTER_WELCOME.format(
                            port=port, domain=domain, path=path, url=url.group(0)
                        )
                    )
                else:
                    print(line)


if __name__ == "__main__":
    kjupyter()



from test.args.models.basic import basic_attr, basic_dict
from test.args.models.no_defaults import no_default_dict, no_default_attr
from test.args.models.positional_and_shape import pos_and_shape

models = [basic_attr, basic_dict, no_default_attr, no_default_dict, pos_and_shape]



from __future__ import absolute_import

from rich.theme import Theme

default = Theme(
    {
        "heading": "green bold",
        "hot": "deep_pink3",
        "grey": "grey54",
        "syntax_col": "cyan bold",
        "examples_col": "grey54",
        "default_col": "deep_pink3",
        "values_col": "italic",
    }
)



from __future__ import absolute_import

import json
import os
import site
import sys
from contextlib import closing
from pathlib import Path
from typing import Optional
from urllib.request import Request, urlopen

from kslurm.installer.version import FlexVersion


def data_dir(home_dir_var: str) -> Path:
    """Get directory to store app data and virtual env

    Returns the data directory for the app to be installed. It first checks if the
    python executable is inside the venv to be upgraded. Then, it checks the home_dir
    environment variable is set. It then checks the XDG_DATA_HOME environment var, then,
    if everything else is empty, returns the default path.

    Args:
        home_dir_var (str): Name of environment variable used to store the home dir path

    Returns:
        Path: Path of the home directory
    """

    dir = (Path(sys.executable) / "../../..").resolve()
    # Dir should have VERSION file
    if (dir / "VERSION").exists():
        return dir

    # If not, check if they have their HOME_DIR set
    if os.getenv(home_dir_var):
        return Path(os.getenv(home_dir_var)).expanduser()  # type: ignore

    # If still nothing, we'll just install at the usual place
    path = os.getenv("XDG_DATA_HOME", Path.home() / ".local/share")
    path = Path(path) / "kutils"

    return path


def bin_dir(home_dir_var: str) -> Path:
    """Get path of directory holding executable files.

    Returns the directory containing the user's local executable files. If the user
    has the HOME_DIR environment variable set, it will return the bin dir within that
    folder. The user will be responsible for ensuring that folder is on the PATH.
    Otherwise, it returns the default getuserbase() bin dir (.local/bin on linux)

    Args:
        home_dir_var (str): Name of environment variable containing HOMEDIR

    Returns:
        Path: Path to folder containing local executable files
    """
    if os.getenv(home_dir_var):
        return Path(os.getenv(home_dir_var), "bin").expanduser()  # type: ignore

    user_base = site.getuserbase()

    bin_dir = os.path.join(user_base, "bin")

    return Path(bin_dir)


def get(url: str):
    """Make an HTTP request and read the response.

    Args:
        url (str): URL to request

    Returns:
        str: Response from the http request read
    """
    request = Request(url, headers={"User-Agent": "Python kslurm"})

    with closing(urlopen(request)) as r:
        return r.read()


def get_version(
    requested_version: Optional[str],
    preview: bool,
    metadata_url: str,
):
    """Retrieves version information and returns a valid version for installation.

    If no specific version is requested, it will return the latest version. Otherwise,
    it will check to make sure the requested version is valid.

    Args:
        requested_version (str or None): Specific version to request. If None, request
            latest version.
        preview (bool): Set to True to allow preview or development versions
        metadata_url (str): url to pipy repository metadata

    Returns:
        Optional[str]: Valid version for installation. Returns None if requested
            version does not exist
    """

    metadata = json.loads(get(metadata_url).decode())

    releases = sorted([FlexVersion.parse(k) for k in metadata["releases"].keys()])

    if requested_version:
        version = FlexVersion.parse(requested_version)
        for release in releases:
            if version == release:
                return release.raw_value
        print(f"Version {requested_version} does not exist.")
        return None
    else:
        for release in reversed(releases):
            if release.prerelease and not preview:
                continue
            return release.raw_value


def get_current_version(datadir: Path) -> Optional[str]:
    """Retrieves current version from the app data directory, if it exists

    Args:
        datadir (Path): Path of the data directory

    Returns:
        Optional[str]: The current version, otherwise None
    """
    if datadir.joinpath("VERSION").exists():
        return datadir.joinpath("VERSION").read_text().strip()
    return None



from __future__ import absolute_import

import itertools as it
from collections import defaultdict
from typing import Any, DefaultDict, Dict, Iterable, List, TypeVar, cast

from kslurm.args.arg_types import (
    Arg,
    FlagArg,
    KeywordArg,
    PositionalArg,
    ShapeArg,
    TailArg,
)
from kslurm.exceptions import CommandLineError

T = TypeVar("T")
S = TypeVar("S", bound=Arg[Any])


def group_by_type(items: Iterable[T]) -> Dict[type, List[T]]:
    groupedDict = cast(DefaultDict[type, List[T]], defaultdict(list))
    for key, value in it.groupby(items, type):
        groupedDict[key] += list(value)
    return groupedDict


class ArgSorter:
    def __init__(self, args: Iterable[Arg[Any]]):
        self.args = list(args)
        self.grouped = group_by_type(self.args)

    @property
    def shapes(self):
        return cast(List[ShapeArg[Any]], self.grouped.get(ShapeArg, []))

    @property
    def keywords(self):
        return cast(List[KeywordArg[Any]], self.grouped.get(KeywordArg, []))

    @property
    def flags(self):
        return cast(List[FlagArg], self.grouped.get(FlagArg, []))

    @property
    def tail(self):
        tail = cast(List[TailArg], self.grouped.get(TailArg, []))
        if len(tail) > 1:
            raise CommandLineError("More than 1 TailArgs provided:" f"{tail}")
        if not tail:
            return None
        return tail[0]

    @property
    def dynamic_args(self):
        return list(it.chain(self.shapes, self.keywords, self.flags))

    @property
    def static_args(self):
        return cast(
            Iterable[PositionalArg[Any]],
            filter(lambda x: isinstance(x, PositionalArg), self.args),
        )

    @property
    def validated_args(self):
        return list(it.chain(self.static_args, self.keywords))



from __future__ import absolute_import

import os
import sys
from typing import List, Union

import kslurm.appconfig as appconfig
import kslurm.models.job_templates as templates
import kslurm.slurm.helpers as helpers
from kslurm.args import ShapeArg
from kslurm.exceptions import TemplateError, ValidationError
from kslurm.models import SlurmModel


class SlurmCommand:
    def __init__(self, args: Union[SlurmModel, TemplateError]):
        self._name = ""
        self._output = ""

        if isinstance(args, TemplateError):
            print(args.msg)
            print("Choose from the following list of templates:\n")
            templates.list_templates()
            exit()

        if args.list_job_templates.value and not args.help.value:
            templates.list_templates()
            exit()

        # set_template returns templated values only if a template is passed
        # if we pass a blank string, the models are returned unchanged
        template = args.job_template.values[0] if args.job_template.value else ""
        template_vals = templates.set_template(
            template, mem=args.mem, cpu=args.cpu, time=args.time
        )

        # Start by setting these three with model/template
        self.time = template_vals.time
        self.cpu = template_vals.cpu
        self.mem = template_vals.mem

        # Then update if values were specifically supplied on the command line
        if args.time.updated:
            self.time = args.time
        if args.cpu.updated:
            self.cpu = args.cpu
        if args.mem.updated:
            self.mem = args.mem

        self.gpu = bool(args.gpu.value)
        self.x11 = bool(args.x11.value)
        if not args.account.values:
            self.account = appconfig.get_config("account")
            if not self.account:
                print(
                    "Account must either be specified using --account, or provided in "
                    "config. A default account can be added to config by running "
                    "kslurm config account <account_name>"
                )
                sys.exit(1)
        else:
            self.account = str(args.account)
        self.cwd = args.directory
        self.test = bool(args.test.value)
        self.job_template = args.job_template
        self._command = args.tail.values
        self.help = bool(args.help.value)

        os.chdir(self.cwd.value)  # type: ignore

        self.script = [self.command]

        self.args = args

    ###
    # Job Paramaters
    ###
    @property
    def time(self):
        return helpers.slurm_time_format(self._time.value)  # type: ignore

    @time.setter
    def time(self, time: ShapeArg[int]):
        self._time = time

    @property
    def name(self):
        if not self._name:
            return self._command[0]
        else:
            return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def output(self):
        return self._output

    @output.setter
    def output(self, output: str):
        self._output = f'--output="{output}"'

    ###
    # Command line strings
    ###
    @property
    def slurm_args(self):
        s = (
            f"--account={self.account} --time={self.time} "
            f"--cpus-per-task={self.cpu} --mem={self.mem}"
        )
        if self.gpu:
            s += " --gres=gpu:1"
        if self.x11:
            s += " --x11"
        return s

    @property
    def command(self):
        return " ".join(self._command)

    @command.setter
    def command(self, commands: List[str]):
        self._command = commands

    @property
    def script(self):
        return self._script

    @script.setter
    def script(self, command: List[str]):
        self._script = "\n".join(["#!/bin/bash"] + command)

    ###
    # Job submission commands
    ###
    @property
    def run(self):
        if self.command:
            return f"echo '{self.command}' | srun {self.slurm_args} bash"
        else:
            return f"salloc {self.slurm_args}"

    @property
    def batch(self):
        if self.test:
            s = "cat"
        else:
            s = (
                f"sbatch {self.slurm_args} --job-name={self.name} "
                f"--parsable {self.output}"
            )
        if self.command:
            return f"echo '{self.script}' | {s}"
        else:
            raise ValidationError("No command given")



from __future__ import absolute_import

import textwrap
from pathlib import Path
from typing import Any, Iterable, List, Optional, TypeVar, Union, cast

from rich.console import Group
from rich.padding import Padding
from rich.table import Table  # type: ignore (Claims Table is not exported??)
from rich.text import Text

from kslurm.style.console import console

from .arg_types import (
    Arg,
    ChoiceArg,
    FlagArg,
    KeywordArg,
    PositionalArg,
    ShapeArg,
    SubCommand,
    TailArg,
)
from .helpers import get_arg_list, group_by_type


def _syntax_format(syntax: str):
    lines = [f"[cyan bold]{line.strip()}[/]" for line in syntax.split(" | ")]
    return "\n".join(lines)


def _header(header: str):
    return f"[bold green]{header.upper()}[/bold green]"


def _section(header: str, body: Union[Table, str]):
    return (
        Group(Padding.indent(_header(header), 4), Padding.indent(body, 8), "\n")
        if body
        else ""
    )


def print_help(script: str, models: object, script_help: str = "") -> None:
    arg_list = get_arg_list(models)
    grouped = group_by_type(arg_list)
    positionals = cast(List[PositionalArg[Any]], grouped.get(PositionalArg, []))
    shapes = cast(List[ShapeArg[Any]], grouped.get(ShapeArg, []))
    keywords = cast(List[KeywordArg[Any]], grouped.get(KeywordArg, []))
    flags = cast(List[FlagArg], grouped.get(FlagArg, []))
    choices = cast(List[ChoiceArg[Any]], grouped.get(ChoiceArg, []))
    subcommands = cast(List[SubCommand], grouped.get(SubCommand, []))
    tail = cast(List[TailArg], grouped.get(TailArg, []))

    positional_names = [f"<{arg.name}>" for arg in positionals]
    script_name = Path(script).name

    command_line_example = "".join(
        [
            "[b]USAGE:[/] ",
            script_name,
            " [hot]<keywords>[/]" if any([shapes, keywords, flags, choices]) else "",
            " " + " ".join(positional_names) if positionals else "",
            " <command>" if subcommands else "",
            f" [cyan]<{tail[0].name.lower()}...>[/]" if tail else "",
            "\n\n",
        ]
    )

    shape_section = _section("Shape args", _shape_table(shapes))
    keyword_section = _section("keyword args", _keyword_table(keywords))
    positional_section = _section("positional args", _positional_table(positionals))
    flag_section = _section("flag args", _flag_table(flags))
    choice_section = _section("choice args", _choice_table(choices))
    subcommand_section = _section("subcommands", _choice_table(subcommands))
    sections = list(
        filter(
            None,
            [
                "\n",
                command_line_example,
                script_help,
                subcommand_section,
                choice_section,
                positional_section,
                shape_section,
                keyword_section,
                flag_section,
            ],
        )
    )
    console.print(*sections)


T = TypeVar("T", bound=Arg[Any])


def _get_helps(args: List[T]):
    return [Text(textwrap.fill(arg.help, 70)) for arg in args]


def _get_defaults(args: List[Any]):
    return [Text(str(arg), style="default_col") for arg in args]


def _help_table(
    rows: Iterable[Iterable[Union[Text, str]]],
    header: Optional[List[str]] = None,
    right_aligned_cols: int = 1,
):
    t = Table.grid(
        padding=(1, 2),
    )
    for _ in range(right_aligned_cols):
        t.add_column(justify="right")
    if header:
        t.add_row(*header, style="bold")
    for r in rows:
        t.add_row(*list(r))
    return t


def _shape_table(args: List[ShapeArg[Any]]):
    if args:
        names = [Text(arg.name, style="bold") for arg in args]
        syntaxes = [_syntax_format(arg.syntax) for arg in args]
        examples = [Text("\n".join(arg.examples), style="grey") for arg in args]
        eg = ["➔" if example else "" for example in examples]
        defaults = _get_defaults([arg.raw_value for arg in args])
        helps = _get_helps(args)

        header = ["", "Syntax", "", "Examples", "Default", ""]
        body = zip(names, syntaxes, eg, examples, defaults, helps)

        return _help_table(body, header, right_aligned_cols=2)
    else:
        return ""


def _positional_table(args: List[PositionalArg[Any]]):
    if args:
        names = [Text(arg.name, style="bold") for arg in args]
        defaults = _get_defaults([arg.raw_value for arg in args])
        helps = _get_helps(args)
        body = zip(names, defaults, helps)
        header = ["", "Default", ""]
        return _help_table(body, header)
    else:
        return ""


def _keyword_table(args: List[KeywordArg[Any]]):
    if args:
        names = [Text(", ".join(arg.match_list), style="bold") for arg in args]
        value_names = [f"<[i]{arg.values_name}[/]>" for arg in args]
        defaults = _get_defaults([arg.values for arg in args])
        helps = _get_helps(args)

        body = zip(names, value_names, defaults, helps)
        header = ["", "", "Default", ""]
        return _help_table(body, header)
    else:
        return ""


def _flag_table(args: List[FlagArg]):
    if args:
        names = [Text(", ".join(arg.match_list), style="bold") for arg in args]
        helps = _get_helps(args)

        body = zip(names, helps)
        return _help_table(body)
    else:
        return ""


S = TypeVar("S", bound=ChoiceArg[Any])


def _choice_table(args: List[S]):
    if args:
        names = [Text(arg.name, style="bold") for arg in args]
        choices = [Text(", ".join(arg.match_list), style="bold") for arg in args]
        helps = _get_helps(args)
        defaults = _get_defaults([arg.raw_value for arg in args])

        body = zip(names, choices, defaults, helps)
        return _help_table(body)
    else:
        return ""



from kslurm.cli.config import (
    ConfigModel,
    config,
)
from kslurm.cli.kbatch import (
    kbatch,
)
from kslurm.cli.kjupyter import (
    kjupyter,
)
from kslurm.cli.kpy import (
    ActivateModel,
    KpyModel,
    MissingPipdirError,
    MissingSlurmTmpdirError,
    VenvCache,
    get_hash,
    kpy,
    pip_freeze,
)
from kslurm.cli.krun import (
    krun,
)
from kslurm.cli.main import (
    ENTRYPOINTS,
    HOME_DIR,
    KslurmModel,
    NAME,
    main,
)

__all__ = [
    "ActivateModel",
    "ConfigModel",
    "ENTRYPOINTS",
    "HOME_DIR",
    "KpyModel",
    "KslurmModel",
    "MissingPipdirError",
    "MissingSlurmTmpdirError",
    "NAME",
    "VenvCache",
    "config",
    "get_hash",
    "kbatch",
    "kjupyter",
    "kpy",
    "krun",
    "main",
    "pip_freeze",
]



from __future__ import absolute_import

from typing import Any, List, NamedTuple

import pytest
from _pytest.fixtures import SubRequest

import kslurm.args.helpers as helpers
import kslurm.args.parser as sc
from kslurm.args import Arg
from kslurm.exceptions import CommandLineErrorGroup

from .models import models
from .models.common import ModelTest


class ModelCase(NamedTuple):
    model: object
    test: List[str]
    outcome: List[Arg[Any]]


@pytest.fixture(scope="module", params=models)
def model(request: SubRequest):  # type: ignore
    return request.param  # type: ignore


def test_relabelled_args_convert_back_to_model(model: ModelTest):
    labelled = helpers.get_arg_list(model.model)
    remodelled = helpers.update_model(labelled, model.model)
    assert remodelled == model.model


def test_parse_args(model: ModelTest):
    for args, outcome in zip(model.tests, model.expected_outcomes):
        if isinstance(outcome[0], Exception):
            try:
                sc.parse_args(args, model.model, True)
            except CommandLineErrorGroup as err_group:
                assert [err.msg for err in outcome] == [  # type: ignore
                    err.msg for err in err_group.sub_errors
                ]
            else:
                assert False
        else:
            parsed_model = sc.parse_args(args, model.model, True)
            labelled_list = helpers.get_arg_list(parsed_model)

            assert set(labelled_list) == set(outcome)  # type: ignore



from __future__ import absolute_import

import inspect
import sys
from typing import Any, Callable, List, Union

import attr

from kslurm.args.parser import parse_args
from kslurm.args.types import WrappedCommand
from kslurm.exceptions import CommandLineError

_CommandFunc = Union[Callable[[Any], None], Callable[[], None]]


def command(func: _CommandFunc) -> WrappedCommand:
    @attr.frozen
    class BlankModel:
        pass

    if not callable(func):
        raise CommandLineError(f"{func} is not callable")

    params = inspect.signature(func)
    exceptions: Any = tuple()
    model = BlankModel
    if len(params.parameters) > 1:
        raise CommandLineError(
            f"{func} must have a single, named arg. Currently: "
            f"{inspect.signature(func)}"
        )
    if len(params.parameters):
        param = next(iter(params.parameters.values()))
        if param.annotation == param.empty:
            raise CommandLineError(f"parameter '{param}' in {func} must be annotated")
        model = param.annotation
        if isinstance(model, str):
            model = eval(model, func.__globals__)

        types = getattr(model, "__args__", None)
        if types is not None:
            model = types[0]
            exceptions = types[1:]

        if not isinstance(model, type):
            raise CommandLineError(
                f"Annotation of {param} in {func} must refer to a class type "
                f"(currently {type(model)})"
            )

    def wrapper(argv: List[str] = sys.argv):
        doc = func.__doc__
        if doc is None:
            doc = ""

        try:
            parsed = parse_args(argv[1:], model(), script_name=argv[0], docstring=doc)
        except exceptions as err:
            parsed = err
        try:
            func(parsed)  # type: ignore
        except TypeError:
            func()  # type: ignore

    return wrapper



from __future__ import absolute_import

from colorama import Fore

SETTINGS_HEADER = f"{Fore.LIGHTBLUE_EX}SETTINGS{Fore.RESET}"
COMMAND_HEADER = f"{Fore.LIGHTBLUE_EX}COMMAND{Fore.RESET}"

KBATCH_MSG = f"""
    {Fore.GREEN}Scheduling Batch Command
        {SETTINGS_HEADER}
            {Fore.WHITE}{{slurm_args}}
        {COMMAND_HEADER}
            {Fore.WHITE}{{command}}
"""

KRUN_CMD_MESSAGE = f""" \
    {Fore.GREEN}Running job
        {SETTINGS_HEADER}
            {Fore.WHITE} {{args}}
        {COMMAND_HEADER}
            {{command}}
"""

INTERACTIVE_MSG = f"""
    {Fore.GREEN}Running interactive session
        {SETTINGS_HEADER}
            {Fore.WHITE} {{args}}
"""

JUPYTER_WELCOME = """
    [hot]Started Jupyter Server!![/]

    [heading]Web browser access:[/]
        You'll need to set up an ssh tunnel. Open a new console
        and enter the tunnel script below.
        Then paste the browser url into your browser of choice!

    [heading]VS Code access:[/]
        Start a VS Code SSH session (see here for more info:
        https://code.visualstudio.com/docs/remote/ssh). Open your
        Jupyter notebook file, then set the Jupyter server to remote:
            Command Pallete > "Jupyter: Specify local or remote Jupyter server for
            connections"
        When prompted, enter the Server URL

    [heading]Tunnel script (replace <address> with your own address):[/]
        ssh -L {port}:{domain}:{port} <username@sub.domain.ext>
    [heading]Browser URL:[/]
        http://localhost:{port}{path}
    [heading]Server URL:[/]
        {url}

    Press Ctrl+C to exit the server
"""



from __future__ import absolute_import, annotations

from typing import List, TypeVar

T = TypeVar("T", bound="CommandLineError")


class CommandLineError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)
        self.msg = msg


class ValidationError(CommandLineError):
    pass


class TailError(CommandLineError):
    pass


class MandatoryArgError(CommandLineError):
    pass


class CommandLineErrorGroup(CommandLineError):
    def __init__(self, msg: str, sub_errors: List[T] = []):
        super().__init__(msg)
        self.sub_errors = sub_errors


class TemplateError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)
        self.msg = msg



from __future__ import absolute_import

import re
from test.args.models import formatters
from test.args.models.common import ModelTest, update_model
from typing import Any, List

import attr

from kslurm.args.arg_types import FlagArg, KeywordArg, PositionalArg, ShapeArg, TailArg


def get_tests(model: object) -> List[List[List[Any]]]:
    return [
        [
            [
                "template",
                "parcour",
                "positional",
                "2",
                "positional2",
                "positional3",
            ]
        ],
        [
            update_model(
                model,
                [
                    "03:00",
                    None,
                    "2",
                    [True, ["parcour"]],
                    "positional",
                    "positional2",
                    "positional3",
                    [None, []],
                ],
            ),
        ],
    ]


@attr.s(auto_attribs=True)
class PositionalAndShapeModel:
    time: ShapeArg[str] = ShapeArg(
        id="random",
        match=lambda arg: bool(re.match(r"^([0-9]{1,2}-)?[0-9]{1,2}:[0-9]{2}$", arg)),
        format=formatters.time,
        value="03:00",
    )

    flag: FlagArg = FlagArg(match=["flag"])

    cpu: ShapeArg[str] = ShapeArg(
        match=lambda arg: bool(re.match(r"^[0-9]+$", arg)), value="1"
    )

    template: KeywordArg[str] = KeywordArg(match=["-j", "template"], num=1, value=False)

    positional: PositionalArg[str] = PositionalArg(id="positional", value="pos1")

    positional2: PositionalArg[str] = PositionalArg(id="positional2", value="pos2")

    positional3: PositionalArg[str] = PositionalArg(id="positional3", value="pos3")

    tail: TailArg = TailArg()


pos_and_shape = ModelTest(
    PositionalAndShapeModel(), *get_tests(PositionalAndShapeModel())
)



from __future__ import absolute_import

import re
from typing import Union

from kslurm.args.arg_types import FlagArg, KeywordArg, PositionalArg, ShapeArg, TailArg


def time_format(time: str):
    if "-" in time:
        days, hhmm = time.split("-")
        hours, min = hhmm.split(":")
    else:
        days = 0
        hours, min = time.split(":")
    return str(int(min) + int(hours) * 60 + int(days) * 60 * 24)


class Templates:
    time: ShapeArg[str] = ShapeArg(
        id="random",
        match=lambda arg: bool(re.match(r"^([0-9]{1,2}-)?[0-9]{1,2}:[0-9]{2}$", arg)),
        format=time_format,
        value="03:00",
    )

    gpu: FlagArg = FlagArg(match=["gpu"])

    number: ShapeArg[int] = ShapeArg(
        match=lambda arg: bool(re.match(r"^[0-9]+$", arg)), value="1", format=int
    )

    job_template: KeywordArg[str] = KeywordArg[str](
        id="junk", match=["-j", "--job-template"], num=1, value=False
    )

    length_5_keyword: KeywordArg[str] = KeywordArg[str](
        id="length_5_keyword",
        match=["--length_5_keyword"],
        num=5,
        value=False,
    )

    lazy_inf_keyword: KeywordArg[str] = KeywordArg[str](
        id="lazy_inf_keyword",
        match=["lazy_inf_keyword"],
        num=-1,
        value=False,
        lazy=True,
    )

    greedy_inf_keyword: KeywordArg[Union[str, bool]] = KeywordArg[Union[str, bool]](
        id="greedy_inf_keyword",
        match=["greedy_inf_keyword"],
        num=-1,
        value=False,
    )

    positional: PositionalArg[str] = PositionalArg(id="positional", value="pos1")

    positional2: PositionalArg[str] = PositionalArg(id="positional2", value="pos2")

    positional3: PositionalArg[str] = PositionalArg(id="positional3", value="pos3")

    positional4: PositionalArg[str] = PositionalArg(id="positional4", value="pos4")

    positional5: PositionalArg[str] = PositionalArg(id="positional5", value="pos5")

    tail: TailArg = TailArg()



from __future__ import absolute_import

import re
import test.args.models.formatters as formatters
from test.args.models.common import ModelTest, update_model
from typing import Any, List

import attr
from typing_extensions import TypedDict

from kslurm.args.arg_types import FlagArg, KeywordArg, ShapeArg, TailArg
from kslurm.exceptions import MandatoryArgError


def get_tests(model: object) -> List[List[List[Any]]]:
    return [
        [
            # Providing everything works
            [
                "07:23",
                "gpu",
                "3",
                "--length_5_keyword",
                "one",
                "of",
                "the",
                "five",
                "values",
                "command",
            ],
            # Raise Exception if missing stuff
            ["gpu"],
        ],
        [
            update_model(
                model,
                [
                    "07:23",
                    True,
                    3,
                    [True, ["one", "of", "the", "five", "values"]],
                    [None, ["command"]],
                ],
            ),
            [
                MandatoryArgError("time has not been provided a value."),
                MandatoryArgError("number has not been provided a value."),
                MandatoryArgError("--length_5_keyword has not been provided a value."),
            ],
        ],
    ]


@attr.s(auto_attribs=True)
class AttrModel:
    time: ShapeArg[str] = ShapeArg(
        id="random",
        match=lambda arg: bool(re.match(r"^([0-9]{1,2}-)?[0-9]{1,2}:[0-9]{2}$", arg)),
        format=formatters.time,
    )

    gpu: FlagArg = FlagArg(match=["gpu"])

    number: ShapeArg[int] = ShapeArg(
        match=lambda arg: bool(re.match(r"^[0-9]+$", arg)), format=int
    )

    length_5_keyword: KeywordArg[str] = KeywordArg(
        id="length_5_keyword",
        match=["--length_5_keyword"],
        num=5,
        value=None,
    )

    tail: TailArg = TailArg()


class TypedDictModel(TypedDict):
    time: ShapeArg[str]
    gpu: FlagArg
    number: ShapeArg[str]
    length_5_keyword: KeywordArg[str]
    tail: TailArg


attrmodel = AttrModel()

typed_dict_model = TypedDictModel(
    time=ShapeArg(
        id="random",
        match=lambda arg: bool(re.match(r"^([0-9]{1,2}-)?[0-9]{1,2}:[0-9]{2}$", arg)),
        format=formatters.time,
    ),
    gpu=FlagArg(match=["gpu"]),
    number=ShapeArg(match=lambda arg: bool(re.match(r"^[0-9]+$", arg))),
    length_5_keyword=KeywordArg[str](
        id="length_5_keyword",
        match=["--length_5_keyword"],
        num=5,
        value=None,
    ),
    tail=TailArg(),
)

no_default_attr = ModelTest(attrmodel, *get_tests(attrmodel))

no_default_dict = ModelTest(typed_dict_model, *get_tests(typed_dict_model))



__submodules__ = ["helpers", "slurm_command"]

__ignore__ = ["T"]

# <AUTOGEN_INIT>
from kslurm.slurm.helpers import (
    div_remainder,
    slurm_time_format,
)
from kslurm.slurm.slurm_command import (
    SlurmCommand,
)

__all__ = ["SlurmCommand", "div_remainder", "slurm_time_format"]

# </AUTOGEN_INIT>



from __future__ import absolute_import

from typing import Tuple


def div_remainder(num: int, denom: int) -> Tuple[int, int]:
    assert denom != 0
    remainder = num % denom
    return (num - remainder) // denom, remainder


def slurm_time_format(min: int) -> str:
    hr, ex_min = div_remainder(min, 60)
    days, ex_hr = div_remainder(hr, 24)
    if days:
        return f"{days}-{ex_hr:02d}:{ex_min:02d}:00"
    else:
        return f"{ex_hr:02d}:{ex_min:02d}:00"



from __future__ import absolute_import

from kslurm.exceptions import TemplateError
from kslurm.models.job_templates import templates


def job_template(arg: str) -> str:
    if arg in templates():
        return arg
    raise TemplateError(f"{arg} is not a valid job-template")



from __future__ import absolute_import

# import itertools as it
# import subprocess
# import sys
# from pathlib import Path

# import attr
# from colorama import Fore, Style

# from kslurm.args.arg_types import PositionalArg
# from kslurm.exceptions import ValidationError
# from kslurm.models import SlurmModel
# from kslurm.slurm import SlurmCommand


# # Helper function for formatting the list of settings in our output
# def setting_list(name: str, setting: str) -> str:
#     return (
#         Fore.YELLOW
#         + name
#         + ": "
#         + Fore.WHITE
#         + Style.BRIGHT
#         + setting
#         + Style.RESET_ALL
#     )


# def profile_validator(profile: str) -> str:
#     profile_path = Path.home() / ".config" / "snakemake"
#     configfiles = [profile_path.glob(f"*/config.{ext}") for ext in ["yml", "yaml"]]
#     profiles = [configfile.parent.name for configfile in it.chain(*configfiles)]
#     if profile in profiles:
#         return profile
#     if profiles:
#         profiles = "\n".join(profiles)
#         profilelist = f"Found the following profiles:\n{profiles}"
#     else:
#         profilelist = f"Did not find any valid profiles in {profile_path}"
#     raise ValidationError(
#         f'{Fore.RED}"{Fore.LIGHTRED_EX + profile + Fore.RED}" '
#         f"is not a valid profile.{Fore.RESET} \n\n{profilelist}"
#     )


# # Extended Model
# @attr.s(auto_attribs=True)
# class SSnakeModel(SlurmModel):
#     profile: PositionalArg[str] = PositionalArg(validator=profile_validator)


# def main():
#     models = SSnakeModel()
#     models.cpu.value = "2"
#     models.profile
#     slurm = SlurmCommand(sys.argv[1:], models)

#     # Get the profile
#     profile = slurm.args.profile.value

#     # Use parent directory name as the job name
#     slurm.name = Path.cwd().name
#     slurm.output = "snakemake-%j.out"

#     # Update our submission script
#     slurm.script = [
#         "source $SNAKEMAKE_VENV_DIR/activate",
#         "panoptes --ip $(hostname -f) 1> panoptes.out 2>&1 &",
#         "PANOPTES_PID=$!",
#         '(tail -F panoptes.out & ) | grep -q "Running on"',
#         "hostname -f",
#         'snakemake --wms-monitor "http://$(hostname -f):5000" '
#         f"--profile {profile} {slurm.command}",
#         "kill $PANOPTES_PID",
#         "rm panoptes.out",
#     ]

#     # Run the process and collect the jobid output.
#     output = subprocess.run(
#         slurm.batch, shell=True, capture_output=True
#     ).stdout.decode()

#     if slurm.test:
#         # output will be the issued command, so we print it
#         print(Fore.WHITE + output)
#     else:
#         # We subtract the last 2 characters of the output
#         # to remove the final "\n" characters and get the
#         # job_id
#         slurmid = output[:-2]

#         # Print a helpful confirmation message
#         print(
#             f"""
#     {Fore.GREEN}Scheduling Snakemake
#         {Fore.LIGHTBLUE_EX}SETTINGS
#             {Fore.WHITE}{slurm.slurm_args}

#             {setting_list("profile", profile)}
#             {setting_list("job_name", slurm.name)}
#             {setting_list("job_id", slurmid)}
#             {setting_list("other_args", slurm.command)}

#     To cancel the job, run:
#         scancel {slurmid}
#         """
#         )


# if __name__ == "__main__":
#     main()



from __future__ import absolute_import

import re


def time(time: str):
    if ":" in time:
        if "-" in time:
            days, hhmm = time.split("-")
            hours, min = hhmm.split(":")
        else:
            days = "0"
            hours, min = time.split(":")
    else:
        raise TypeError(
            f'Invalid format for time: "{time}"\n'
            f"Must be as [xx-]xx:xx or x where x is a number"
        )
    return int(min) + int(hours) * 60 + int(days) * 60 * 24


def mem(mem: str):
    match = re.match(r"^[0-9]+", mem)
    if match:
        num = int(match.group())
    else:
        raise Exception("Memory is not formatted correctly")
    if "G" in mem:
        return num * 1000
    else:
        return num



from __future__ import absolute_import

import subprocess
from typing import Union

from colorama import Fore

import kslurm.text as txt
from kslurm.args.command import command
from kslurm.exceptions import TemplateError
from kslurm.models import SlurmModel
from kslurm.slurm import SlurmCommand


@command
def kbatch(args: Union[SlurmModel, TemplateError]):
    """Submit a job using sbatch

    Supports scripts (e.g. ./script.sh) or direct commands (e.g. cp dir/file.txt dir2)

    When your command contains bash interpreted elements such as $VARIABLES and
    $(subshells), these will be immediately expanded. Normally, this behaviour
    is fine, but sometimes they should only be interpretted on the allocated cluster.
    For instance, $SLURM_TMPDIR will evaluate to "" unless it is interpretted later.
    To force this behaviour, wrap the $VARIABLE or $(subshell) in quotes:
        '$SLURM_TMPDIR'
        '$(hostname)'
    """

    slurm = SlurmCommand(args)
    command = slurm.command if slurm.command else f"{Fore.RED}Must provide a command"

    print(txt.KBATCH_MSG.format(slurm_args=slurm.slurm_args, command=command))
    if slurm.command:
        proc = subprocess.run(
            slurm.batch, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )

        out = proc.stdout.decode()

        if proc.returncode != 0:
            print(Fore.WHITE + out)
        if slurm.test:
            # output will be the issued command, so we print it
            print(Fore.WHITE + out)
        else:
            # We subtract the last character of the output
            # to remove the final "\n" character and get the
            # job_id
            slurmid = out[:-1]

            print(
                f"""Scheduled job {slurmid}.
    To cancel, run:
        scancel {slurmid}
        """
            )


if __name__ == "__main__":
    kbatch()



from __future__ import absolute_import

import functools as ft
import itertools as it
import json
from pathlib import Path
from typing import Dict, cast

import attr
from tabulate import tabulate
from typing_extensions import TypedDict

from kslurm.args import ShapeArg


class Templates(TypedDict):
    cpus: str
    mem: str
    time: str


@attr.s(auto_attribs=True)
class TemplateArgs:
    mem: ShapeArg[int]
    cpu: ShapeArg[int]
    time: ShapeArg[int]


@ft.lru_cache(None)
def templates() -> Dict[str, Templates]:
    file = Path(f"{__file__}/../../data/slurm_job_templates.json").resolve()

    if not file.exists():
        raise FileNotFoundError(
            "Could not locate slurm_job_templates.json " f"at {file.resolve()}"
        )
    with file.open() as data_stream:
        data = json.load(data_stream)
    cast(Dict[str, Templates], data)
    return data


def set_template(
    template: str, mem: ShapeArg[int], cpu: ShapeArg[int], time: ShapeArg[int]
):
    if template:
        if template not in templates():
            raise Exception(f"{template} is not a valid template")

        return TemplateArgs(
            mem=mem.set_value(templates()[template]["mem"]),
            cpu=cpu.set_value(templates()[template]["cpus"]),
            time=time.set_value(templates()[template]["time"]),
        )
    else:
        return TemplateArgs(mem=mem, cpu=cpu, time=time)


def list_templates():
    labelled_values = list(templates().values())
    headers = ["name"] + list(labelled_values[0].keys())
    values = [list(value.values()) for value in labelled_values]
    entries = list(zip(templates().keys(), values))
    table = [list(it.chain([entry[0]], entry[1])) for entry in entries]
    print(tabulate(table, headers=headers, tablefmt="presto"))



__submodules__ = ["arg_types", "help", "parser", "arg_sorter"]

__ignore__ = ["T", "S"]

# <AUTOGEN_INIT>
from kslurm.args.arg_types import (
    Arg,
    ChoiceArg,
    FlagArg,
    KeywordArg,
    PositionalArg,
    ShapeArg,
    SubCommand,
    TailArg,
)
from kslurm.args.help import (
    print_help,
)
from kslurm.args.parser import (
    parse_args,
)
from kslurm.args.arg_sorter import (
    ArgSorter,
    group_by_type,
)

__all__ = [
    "Arg",
    "ArgSorter",
    "ChoiceArg",
    "FlagArg",
    "KeywordArg",
    "PositionalArg",
    "ShapeArg",
    "SubCommand",
    "TailArg",
    "group_by_type",
    "parse_args",
    "print_help",
]

# </AUTOGEN_INIT>



from __future__ import absolute_import

import subprocess
from pathlib import Path
from typing import List
from venv import EnvBuilder

from kslurm.args import parse_args
from kslurm.installer.utils import bin_dir, data_dir, get_current_version, get_version
from kslurm.installer.version import FlexVersion
from kslurm.models.update import UpdateModel
from kslurm.style import console

METADATA_URL = "https://pypi.org/pypi/kslurm/json"


def install(args: List[str], name: str, home_dir: str, entrypoints: List[str] = []):
    console.print(
        "kslurm update is deprecated. If you haven't already, please switch to a pipx "
        "based install (follow the instructions on the README: "
        "https://github.com/pvandyken/kslurm)"
    )
    parsed = parse_args(args, UpdateModel())
    data = data_dir(home_dir)
    bin = bin_dir(home_dir)
    print("Checking for Updates")
    version = get_version(parsed.version.value, False, METADATA_URL)
    if version is None:
        return 1
    try:
        if FlexVersion.parse(version) == get_current_version(data):
            print(f"Already up to date! (v{version})")
            return 0
    except ValueError:
        # Can't compare versions, so we update
        pass

    specification = f"{name}=={version}"

    return run_installation(version, specification, data, bin, True, entrypoints)


def run_installation(
    version: str,
    specification: str,
    data_dir: Path,
    bin_dir: Path,
    update: bool = False,
    entrypoints: List[str] = [],
):
    """
    Installs Software.
    """
    try:
        data_dir.mkdir(parents=True, exist_ok=True)
        bin_dir.mkdir(parents=True, exist_ok=True)
        env_path = _make_env(data_dir, update)
        _install_library(specification, env_path)
        if entrypoints:
            _make_bin(bin_dir, data_dir, entrypoints)

        data_dir.joinpath("VERSION").write_text(version)

        return 0
    except subprocess.CalledProcessError as e:
        print(f"\nAn error has occurred: {e}\n{e.stderr}")
        return e.returncode


def _make_env(data_dir: Path, update: bool = False) -> Path:
    """Return path to virtualenv, creating one in the process if necessary

    Creates a virtualenv in the directory specified: `datadir/venv/...`. If `update` is
    `True` and the `data_dir` already contains a virtualenv, the path to the virtualenv
    will be returned with no creation

    Args:
        data_dir (Path): Directory in which the venv should be created (or looked for)
        update (bool, optional): If true and a virtualenv exists at `data_dir/venv`, no
            venv will be created. Defaults to False.

    Returns:
        Path: Path of the virtualenv
    """
    env_path = data_dir / "venv"
    if update and _is_venv(env_path):
        return env_path
    print("Making virtual env")
    EnvBuilder(with_pip=True, clear=True).create(str(env_path))
    return env_path


def _is_venv(venv: Path) -> bool:
    python = venv / "bin/python"
    pip = venv / "bin/pip"
    # If pip and python exist, that's enough to install something,
    # so we assume it's a venv
    if python.exists() and pip.exists():
        return True
    return False


def _install_library(specification: str, env_path: Path) -> None:
    """Pip installs the specification using the Python found in the env_path

    Args:
        specification (str): Valid label for pip install (e.g. pkg==x.x.x, git=url,
            path)
        env_path (Path): Path to virtualenv where package should be installed"""
    print("Installing")
    python = env_path.joinpath("bin/python")

    subprocess.run(
        [str(python), "-m", "pip", "install", specification],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        check=True,
    )


def _make_bin(bin_dir: Path, data_dir: Path, entrypoints: List[str]) -> None:
    bin_dir.mkdir(parents=True, exist_ok=True)

    for script in entrypoints:
        target_script = "venv/bin/" + script
        if bin_dir.joinpath(script).is_symlink():
            bin_dir.joinpath(script).unlink()

        bin_dir.joinpath(script).symlink_to(data_dir.joinpath(target_script))



from __future__ import absolute_import, annotations

import importlib.resources as impr
import os
import subprocess as sp
import tempfile
from pathlib import Path

from shellingham import ShellDetectionFailure, detect_shell


class Shell:
    """
    Represents the current shell.
    """

    _shell = None

    def __init__(self, name: str, path: str) -> None:
        self._name = name
        self._path = path

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> str:
        return self._path

    @classmethod
    def get(cls) -> Shell:
        """
        Retrieve the current shell.
        """
        if cls._shell is not None:
            return cls._shell

        try:
            name, path = detect_shell(os.getpid())
        except (RuntimeError, ShellDetectionFailure):
            shell = None

            if os.name == "posix":
                shell = os.environ.get("SHELL")
            elif os.name == "nt":
                shell = os.environ.get("COMSPEC")

            if not shell:
                raise RuntimeError("Unable to detect the current shell.")

            name, path = Path(shell).stem, shell

        if name != "bash":
            print("At this time, only bash shell is supported.")
            exe = sp.run(["command", "-v", "bash"], capture_output=True)
            try:
                exe.check_returncode()
            except sp.CalledProcessError:
                print("No bash executable found on $PATH. Aborting")
            name = "bash"
            path = exe.stdout.decode()

        cls._shell = cls(name, path)

        return cls._shell

    def activate(self, env: Path):
        activate_script = self._get_activate_script()
        activate_path = env / "bin" / activate_script

        if self._name == "bash":
            with impr.path("kslurm.bin", "kpy-init.sh") as path:
                os.environ["activate_path"] = str(activate_path)
                os.environ["kslurm_next_cmd"] = tempfile.mkstemp(prefix="kslurm")[1]
                session = sp.run([self._path, "--init-file", path.resolve()])
            cmd = None
            if session.returncode == 224:
                with open(os.environ["kslurm_next_cmd"]) as f:
                    cmd = f.read().split()
                if cmd[0] != "kpy":
                    cmd = None
            os.remove(os.environ["kslurm_next_cmd"])
            del os.environ["kslurm_next_cmd"]
            del os.environ["activate_path"]
            return cmd

    def _get_activate_script(self) -> str:
        if self._name == "fish":
            suffix = ".fish"
        elif self._name in ("csh", "tcsh"):
            suffix = ".csh"
        elif self._name in ("powershell", "pwsh"):
            suffix = ".ps1"
        elif self._name == "cmd":
            suffix = ".bat"
        else:
            suffix = ""

        return "activate" + suffix

    def _get_source_command(self) -> str:
        if self._name in ("fish", "csh", "tcsh"):
            return "source"
        return "."

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}("{self._name}", "{self._path}")'



from __future__ import absolute_import

from pathlib import Path
from unittest import mock

from pytest import CaptureFixture

from kslurm.cli import kbatch


def test_batch_submits_testmode(capsys: CaptureFixture[str]):
    with mock.patch("subprocess.run") as subprocess:

        kbatch(
            [
                "kbatch",
                "-t",
                "--account",
                "ctb-akhanf",
                "command",
            ]
        )

        out = capsys.readouterr()
        print(out)
        assert (
            "--account=ctb-akhanf --time=03:00:00 --cpus-per-task=1 --mem=4000"
            in str(out)
        )
        subprocess.assert_called_once_with(
            "echo '#!/bin/bash\ncommand' | cat", shell=True, stdout=-1, stderr=-2
        )


def test_params_can_be_altered(capsys: CaptureFixture[str]):
    with mock.patch("subprocess.run") as subprocess:
        starting_cwd = Path.cwd()

        kbatch(
            [
                "kbatch",
                "1-33:11",
                "5G",
                "gpu",
                "--account",
                "some-account",
                "-j",
                "Regular",
                "./test",
                "command",
            ],
        )

        out = capsys.readouterr()
        print(str(out))
        assert (
            "--account=some-account --time=2-09:11:00 --cpus-per-task=8 "
            "--mem=5000 --gres=gpu:1" in str(out)
        )
        subprocess.assert_called_once_with(
            "echo '#!/bin/bash\ncommand' | sbatch --account=some-account "
            "--time=2-09:11:00 --cpus-per-task=8 --mem=5000 --gres=gpu:1 "
            "--job-name=command --parsable ",
            shell=True,
            stdout=-1,
            stderr=-2,
        )
        assert Path.cwd() == starting_cwd / "test"



from __future__ import absolute_import

from typing import List

from typing_extensions import Protocol


class WrappedCommand(Protocol):
    def __call__(self, argv: List[str] = ...) -> None:
        ...



__submodules__ = [
    "formatters",
    "slurm",
    "update",
    "validators",
    "job_templates",
]

# <AUTOGEN_INIT>
from kslurm.models.formatters import (
    mem,
    time,
)
from kslurm.models.slurm import (
    SlurmModel,
)
from kslurm.models.update import (
    UpdateModel,
    VERSION_REGEX,
)
from kslurm.models.validators import (
    job_template,
)
from kslurm.models.job_templates import (
    TemplateArgs,
    Templates,
    list_templates,
    set_template,
    templates,
)

__all__ = [
    "SlurmModel",
    "TemplateArgs",
    "Templates",
    "UpdateModel",
    "VERSION_REGEX",
    "job_template",
    "list_templates",
    "mem",
    "set_template",
    "templates",
    "time",
]

# </AUTOGEN_INIT>



from __future__ import absolute_import, annotations

import re
from typing import Union, cast

import semver


class FlexVersion(semver.VersionInfo):
    raw_value: str
    _REGEX = re.compile(
        r"""
            ^
            (?P<major>0|[1-9]\d*)
            (?:\.
                (?P<minor>0|[1-9]\d*)
            )?
            (?:\.
                (?P<patch>0|[1-9]\d*)
            )?
            [._-]?
            (?:(?P<prerelease>
                (?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)
                (?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*
            ))?
            (?:\+(?P<build>
                [0-9a-zA-Z-]+
                (?:\.[0-9a-zA-Z-]+)*
            ))?
            $
        """,
        re.VERBOSE,
    )

    @classmethod
    def parse(cls, version: Union[str, bytes]):
        parsed = cast("FlexVersion", super().parse(version))
        parsed.raw_value = str(version)
        return parsed



from __future__ import absolute_import

import pytest

import kslurm.models.validators as validators
from kslurm.exceptions import TemplateError


class TestJobTemplateValidator:
    @pytest.mark.parametrize("arg", ["16core64gb24h", "Fat", "Regular"])
    def test_args_that_should_work(self, arg: str):
        assert validators.job_template(arg)

    @pytest.mark.parametrize("arg", ["random", "nonsense", "notfound"])
    def test_args_that_shouldnt_work(self, arg: str):
        with pytest.raises(TemplateError):
            assert not validators.job_template(arg)



from __future__ import absolute_import

import pytest
from hypothesis import given
from hypothesis import strategies as st

from kslurm.models import formatters
from kslurm.slurm.helpers import div_remainder, slurm_time_format


@given(num=st.integers(), denom=st.integers())
def test_div_remainder(num: int, denom: int):
    if denom == 0:
        with pytest.raises(AssertionError):
            div_remainder(num, denom)
    else:
        res, rem = div_remainder(num, denom)
        assert (res * denom) + rem == num


@given(min=st.integers(min_value=0))
def test_slurm_time_format(min: int):
    res = slurm_time_format(min)
    # Strip off last ":00", which won't be recognized by the formatter
    assert formatters.time(res[:-3]) == min



from __future__ import absolute_import

import json
from pathlib import Path
from typing import Dict, Optional

import appdirs

_STATE: Dict[str, Dict[str, str]] = {"config": {}}

CONFIG_PATH = Path(appdirs.user_config_dir("kslurm"), "config.json")


def _load_config():
    if not _STATE["config"]:
        if not CONFIG_PATH.exists():
            CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
            with CONFIG_PATH.open('w') as f:
                json.dump({}, f)
        try:
            with CONFIG_PATH.open("r") as f:
                _STATE["config"] = json.load(f)
        except json.JSONDecodeError:
            print(f"Error decoding config file '{CONFIG_PATH}'")
            exit()


def get_config(entity: str) -> Optional[str]:
    _load_config()
    return _STATE["config"].get(entity, None)


def set_config(entity: str, value: str):
    _load_config()
    _STATE["config"][entity] = value
    with CONFIG_PATH.open("w") as f:
        json.dump(_STATE["config"], f)



# pyright: reportPrivateUsage=false, reportUnknownArgumentType=false
from __future__ import absolute_import

import itertools as it

import pytest
from pytest import fixture

import kslurm.args.parser as sc
from kslurm.args import KeywordArg
from kslurm.args.arg_types import TailArg

from .arg_lists import DummyArgs
from .model_lists import positional_models


@fixture
def models():
    return positional_models(3)


@fixture
def args():
    return DummyArgs()


def test_get_keywork_group_size(args: DummyArgs):
    group_sizes = [sc._get_keyword_group_size(arg) for arg in args.two_keyword]
    assert group_sizes == [0, 0, 1, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 0]


class TestGroupArgs:
    def test_collects_nonkeyword_into_empty_running_list(self, args: DummyArgs):
        arg_list = sc._group_args(args.two_keyword[0:2], [], None)
        assert list(arg_list) == args.two_keyword[0:2]

    def test_merges_args_with_keyword_according_to_group_size(self, args: DummyArgs):
        arg_list = sc._group_args(args.two_keyword[2:4], [], None)
        assert isinstance(args.two_keyword[2], KeywordArg)
        assert list(arg_list) == [
            args.two_keyword[2].add_values([args.two_keyword[3].value])
        ]

        arg_list = sc._group_args(args.two_keyword[5:11], [], None)
        assert isinstance(args.two_keyword[5], KeywordArg)
        assert list(arg_list) == [
            args.two_keyword[5].add_values(
                [model.raw_value for model in args.two_keyword[6:11]]
            )
        ]

    def test_continues_collecting_when_past_keyword_group(self, args: DummyArgs):
        arg_list = sc._group_args(args.two_keyword, [], None)
        assert isinstance(args.two_keyword[2], KeywordArg)
        assert isinstance(args.two_keyword[5], KeywordArg)
        assert list(arg_list) == list(
            it.chain(
                args.two_keyword[0:2],
                [args.two_keyword[2].add_values([args.two_keyword[3].raw_value])],
                args.two_keyword[4:5],
                [
                    args.two_keyword[5].add_values(
                        [model.raw_value for model in args.two_keyword[6:11]]
                    )
                ],
                args.two_keyword[11:],
            )
        )

    @pytest.mark.parametrize("count", [0, 1, 2, 3])
    def test_adds_right_amount_of_positional_args_before_tail(
        self, args: DummyArgs, count: int
    ):
        pos_models = positional_models(count)
        arg_list = sc._group_args(args.positional_patchwork, pos_models, TailArg())
        res = args.positional_patchwork[0 : (count * 2) + 1]
        if count < 3:
            res += [
                TailArg().add_values(
                    [
                        arg.raw_value
                        for arg in args.positional_patchwork[(count * 2) + 1 :]
                    ]
                )
            ]
        arg_l = list(arg_list)
        assert arg_l == res
        if count < 3:
            tail = arg_l[-1]
            assert isinstance(tail, TailArg)
            assert tail.raise_exception is False

    def test_records_exception_if_tail_args_present_but_no_model(self, args: DummyArgs):
        pos_models = positional_models(1)
        arg_list = sc._group_args(args.positional_patchwork, pos_models, None)
        tail = list(arg_list)[-1]
        assert isinstance(tail, TailArg)
        assert tail.raise_exception is True

    def test_does_not_capture_positional_argument_past_keyword_group(
        self, args: DummyArgs
    ):
        pos_models = positional_models(1)
        arg_list = sc._group_args(args.positional_wrapped_keyword, pos_models, None)
        model = args.positional_wrapped_keyword
        assert isinstance(model[2], KeywordArg)
        assert list(arg_list) == list(
            it.chain(
                model[0:2],
                [model[2].add_values([model[3].value])],
                [TailArg().add_values([model[4].raw_value])],
            )
        )

    def test_lazy_inf_keyword_consumes_until_shape_or_keyword(self, args: DummyArgs):
        model = args.lazy_keywords
        pos_models = positional_models(1)
        arg_list = sc._group_args(model, pos_models, None)
        assert isinstance(model[5], KeywordArg)
        assert isinstance(model[2], KeywordArg)
        assert list(arg_list) == list(
            it.chain(
                model[0:2],
                [model[2].add_values([model.value for model in model[3:5]])],
                [
                    model[5].add_values(  # type: ignore
                        [model.value for model in model[6:10]]
                    )
                ],
                model[10:],
            )
        )

    def test_greedy_inf_keyword_consumes_remaining_arguments(self, args: DummyArgs):
        model = args.greedy_keywords
        arg_list = sc._group_args(model, [], None)
        assert isinstance(model[2], KeywordArg)
        assert list(arg_list) == model[0:2] + [
            model[2].add_values([model.raw_value for model in model[3:]])
        ]
